@@ -15,26 +15,27 @@
 #
 
 from __future__ import absolute_import, division, generators, unicode_literals, print_function, nested_scopes, \
-    with_statement 
-
-from builtins import bytes
+    with_statement
 
 import datetime
 import logging as log
+import socket
+import sys
+
+import ipaddress
+from builtins import bytes
+from cryptography import x509
+from cryptography.hazmat.backends import default_backend
+from cryptography.hazmat.primitives import hashes
+from cryptography.hazmat.primitives import serialization
+from cryptography.hazmat.primitives.asymmetric import ec
+from cryptography.hazmat.primitives.asymmetric import rsa
+from cryptography.x509.oid import NameOID, ExtensionOID
 from six import string_types, binary_type
 
 from .errors import VenafiConnectionError, ServerUnexptedBehavior, BadData, ClientBadData
 from .http import HTTPStatus
-
-from cryptography.hazmat.backends import default_backend
-from cryptography.hazmat.primitives import serialization
-from cryptography.hazmat.primitives.asymmetric import rsa
-from cryptography.hazmat.primitives.asymmetric import ec
-from cryptography import x509
-from cryptography.x509.oid import NameOID, ExtensionOID
-from cryptography.hazmat.primitives import hashes
-import ipaddress
-import sys
+from .policy import PolicySpecification
 
 
 MIME_JSON = "application/json"
@@ -42,6 +43,7 @@ MIME_HTML = "text/html"
 MIME_TEXT = "text/plain"
 MIME_CSV = "text/csv"
 MIME_ANY = "*/*"
+LOCALHOST = "127.0.0.1"
 
 
 class CertField:
@@ -61,6 +63,20 @@ def log_errors(data):
         log.error("%s: %s" % (e['code'], e['message']))
 
 
+def get_ip_address():
+    s = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
+    ip = None
+    try:
+        # doesn't even have to be reachable, it just allows the socket to be opened
+        s.connect(('10.255.255.255', 1))
+        ip = s.getsockname()[0]
+    except Exception:
+        ip = LOCALHOST
+    finally:
+        s.close()
+    return ip
+
+
 class KeyType:
     ALLOWED_SIZES = [2048, 3072, 4096, 8192]
     ALLOWED_CURVES = ["p256", "p384", "p521"]
@@ -78,7 +94,8 @@ class KeyType:
                 log.error("unknown size: %s" % option)
                 raise BadData
         elif self.key_type == KeyType.ECDSA:
-            option = {"secp521r1": "p521", "secp384r1": "p384", "secp256r1": "p256", "p256": "p256", "p384":"p384", "p521": "p521"}[option.lower().strip()]
+            option = {"secp521r1": "p521", "secp384r1": "p384", "secp256r1": "p256", "p256": "p256", "p384": "p384",
+                      "p521": "p521"}[option.lower().strip()]
             if option not in KeyType.ALLOWED_CURVES:
                 log.error("unknown curve: %s, should be one of %s" % (option, KeyType.ALLOWED_CURVES))
                 raise BadData
@@ -130,10 +147,14 @@ class TokenInfo:
 
 
 class Policy:
+    # This is the Certificate Issuing Template.
+    # It is used by both TPP and Cloud implementations, hence the name was adapted to something more generic.
     def __init__(self, policy_id=None, company_id=None, name=None, system_generated=None,
                  creation_date=None, subject_cn_regexes=None, subject_o_regexes=None,
                  subject_ou_regexes=None, subject_st_regexes=None, subject_l_regexes=None, subject_c_regexes=None,
-                 san_regexes=None, key_types=None, key_reuse=None):
+                 san_regexes=None, key_types=None, key_reuse=None, cert_authority=None, cert_authority_account_id=None,
+                 cert_authority_product_option_id=None, priority=None, modification_date=None, status=None, reason=None,
+                 validity_period=None, recommended_settings=None):
         """
         :param str policy_id:
         :param str company_id:
@@ -149,6 +170,15 @@ class Policy:
         :param list[str] san_regexes:
         :param list[KeyType] key_types:
         :param bool key_reuse:
+        :param str cert_authority:
+        :param str cert_authority_account_id:
+        :param str cert_authority_product_option_id:
+        :param int priority:
+        :param str modification_date:
+        :param str status:
+        :param str reason:
+        :param str validity_period:
+        :param RecommendedSettings recommended_settings:
         """
         self.id = policy_id
         self.company_id = company_id
@@ -165,6 +195,16 @@ class Policy:
         self.key_types = key_types
         self.key_reuse = key_reuse
 
+        self.cert_authority = cert_authority
+        self.cert_authority_account_id = cert_authority_account_id
+        self.cert_authority_product_option_id = cert_authority_product_option_id
+        self.priority = priority
+        self.modification_date = modification_date
+        self.status = status
+        self.reason = reason
+        self.validity_period = validity_period
+        self.recommended_settings = recommended_settings
+
     def __repr__(self):
         return "Policy:\n" + "\n".join(["  %s: %s" % (k, v) for k, v in (
             ("Id", self.id),
@@ -177,6 +217,27 @@ class Policy:
         return self.name
 
 
+class RecommendedSettings:
+    def __init__(self, subject_o_value=None, subject_ou_value=None, subject_l_value=None, subject_st_value=None,
+                 subject_c_value=None, key_type=None, key_reuse=None):
+        """
+        :param str subject_o_value:
+        :param str subject_ou_value:
+        :param str subject_l_value:
+        :param str subject_st_value:
+        :param str subject_c_value:
+        :param KeyType key_type:
+        :param bool key_reuse:
+        """
+        self.subjectOValue = subject_o_value
+        self.subjectOUValue = subject_ou_value
+        self.subjectLValue = subject_l_value
+        self.subjectSTValue = subject_st_value
+        self.subjectCValue = subject_c_value
+        self.keyType = key_type
+        self.keyReuse = key_reuse
+
+
 class CertificateRequest:
     def __init__(self, cert_id=None,
                  san_dns=None,
@@ -197,7 +258,8 @@ class CertificateRequest:
                  country=None,
                  province=None,
                  locality=None,
-                 origin=None
+                 origin=None,
+                 custom_fields=None,
                  ):
         """
         :param str cert_id: Certificate request id. Generating by server.
@@ -215,6 +277,7 @@ class CertificateRequest:
         :param str common_name: Common name of certificate. Usually domain name.
         :param str thumbprint: Certificate thumbprint. Can be used for identifying certificate on the platform.
         :param str origin: application identifier
+        :param list[CustomField] custom_fields: list of custom fields values to be added to the certificate.
         """
 
         self.chain_option = "last"
@@ -241,6 +304,8 @@ class CertificateRequest:
         # CSR should be last, because it checks subject to match with over parameters
         self.csr = csr
         self.origin = origin or "Venafi VCert-Python"
+        self.custom_fields = custom_fields
+        self.cert_guid = None
 
     def __setattr__(self, key, value):
         if key == "key_password":
@@ -367,26 +432,26 @@ class CertificateRequest:
                 # per https://docs.microsoft.com/en-us/windows/win32/seccertenroll/about-utf8string
                 # Construct the array of bytes (note bytes are strings in python2)
                 # by inserting the header consisting of the tag '0x0C' followed by the length of the upn
-                if (sys.version_info > (3,0)):
+                if sys.version_info > (3, 0):
                     # For python3, since we have native bytes available we'll convert the string into bytes.
                     # However, for cases when this method is called with inputs extracted from existing certificates
                     # & csrs, the input will already be encoded as bytes. So we'll check the input, if it is a 
                     # string we'll convert it into a bytes object then insert our header. Otherwise, we'll just
                     # insert the header in the passed in bytes.
-                    if (isinstance(upn,str)):
+                    if isinstance(upn, str):
                         bupn = bytes(upn,'utf-8')
                     else:
                         bupn = upn
                     values = [12,len(upn)]
                     header = bytes(values)
                     data = header + bupn
-                    alt_names.append(x509.OtherName(UPNOID,data))
+                    alt_names.append(x509.OtherName(UPNOID, data))
                 else:
                     # For python2, since there is no native bytes method, we'll use the bytes method 
                     # in the future module to convert the string we constructed into bytes.
                     bupn = ''.join(chr(x) for x in [12,len(upn)])
                     bupn = bupn + str(upn)
-                    alt_names.append(x509.OtherName(UPNOID,bytes(bupn,'utf-8')))
+                    alt_names.append(x509.OtherName(UPNOID,bytes(bupn, 'utf-8')))
 
         csr_builder = csr_builder.add_extension(
             x509.SubjectAlternativeName(alt_names),
@@ -442,6 +507,32 @@ class CertificateRequest:
             self.key_type = zone.key_type
 
 
+class CustomField:
+    def __init__(self, name, value):
+        """
+        :param str name: The name of the custom field
+        :param str value: The value to be added to the specified custom field
+        """
+        self.name = name
+        self.value = value
+
+    @property
+    def name(self):
+        return self._name
+
+    @name.setter
+    def name(self, value):
+        self._name = value
+
+    @property
+    def value(self):
+        return self._value
+
+    @value.setter
+    def value(self, value):
+        self._value = value
+
+
 class RevocationRequest:
     class RevocationReasons:
         NoReason = 0
@@ -465,7 +556,7 @@ class RevocationRequest:
 
 
 CLIENT_ID = "vcert-sdk"  # type: str
-SCOPE = "certificate:manage,revoke"  # type: str
+SCOPE = "certificate:manage,revoke;configuration:manage"  # type: str
 
 
 class Authentication:
@@ -482,6 +573,32 @@ class Authentication:
         self.state = state
 
 
+class AppDetails:
+    def __init__(self, app_id=None, cit_map=None, company_id=None, name=None, description=None,
+                 owner_ids_and_types=None, fq_dns=None, internal_fq_dns=None, external_ip_ranges=None,
+                 internal_ip_ranges=None, internal_ports=None, fully_qualified_domain_names=None, ip_ranges=None,
+                 ports=None, org_unit_id=None):
+        """
+        :param str app_id:
+        :param dict cit_map:
+        """
+        self.app_id = app_id
+        self.cit_alias_id_map = cit_map
+        self.company_id = company_id
+        self.name = name
+        self.description = description
+        self.owner_ids_and_types = owner_ids_and_types
+        self.fq_dns = fq_dns
+        self.internal_fq_dns = internal_fq_dns
+        self.external_ip_ranges = external_ip_ranges
+        self.internal_ip_ranges = internal_ip_ranges
+        self.internal_ports = internal_ports
+        self.fully_qualified_domain_names = fully_qualified_domain_names
+        self.ip_ranges = ip_ranges
+        self.ports = ports
+        self.org_unit_id = org_unit_id
+
+
 class CommonConnection:
 
     def auth(self):
@@ -493,19 +610,20 @@ class CommonConnection:
 
     def request_cert(self, request, zone):
         """
-        Making request to certificate. It will generate CSR from data if CSR not specified,
-        generate key if required and send to server for signing. Set request.id for retrieving certificate.
+        Making request to certificate. It will generate CSR from data if CSR not specified, generate key if required and send to server for signing. Set request.id for retrieving certificate.
+
         :param CertificateRequest request: Certificate in PEM format
         :param str zone: Venafi zone tag name
-        :rtype bool : Success
+        :rtype: bool
         """
         raise NotImplementedError
 
     def retrieve_cert(self, request):
         """
         Get signed certificate from server by request.id
+
         :param CertificateRequest request:
-        :rtype Certificate
+        :rtype: Certificate
         """
         raise NotImplementedError
 
@@ -525,13 +643,28 @@ class CommonConnection:
     def read_zone_conf(self, tag):
         """
         :param str tag:
-        :rtype ZoneConfig
+        :rtype: ZoneConfig
         """
         raise NotImplementedError
 
     def import_cert(self, request):
         raise NotImplementedError
 
+    def get_policy(self, zone):
+        """
+        :param str zone:
+        :rtype: PolicySpecification
+        """
+        raise NotImplementedError
+
+    def set_policy(self, zone, policy_spec):
+        """
+        :param str zone:
+        :param PolicySpecification policy_spec:
+        :rtype: None
+        """
+        raise NotImplementedError
+
     @staticmethod
     def process_server_response(r):
         if r.status_code not in (HTTPStatus.OK, HTTPStatus.ACCEPTED, HTTPStatus.CREATED, HTTPStatus.CONFLICT):
@@ -539,8 +672,7 @@ class CommonConnection:
                 log_errors(r.json())
             except:
                 log_errors(r.content)
-            raise VenafiConnectionError("Server status: %s\n Response: %s" %
-                                        (r.status_code, r.request.url))
+            raise VenafiConnectionError("Server status: %s\nResponse: %s" % (r.status_code, r.request.url))
 
         content_type = r.headers.get("content-type")
         # Content-type not present, return status and reason (if any)
@@ -554,7 +686,7 @@ class CommonConnection:
             return r.status_code, r.text
         # content-type in response is  application/json; charset=utf-8
         elif content_type.startswith(MIME_JSON):
-            log.debug(r.content.decode())
+            # log.debug(r.content.decode())
             return r.status_code, r.json()
         elif content_type.startswith(MIME_CSV):
             log.debug(r.content.decode())
