@@ -23,17 +23,16 @@ import logging as log
 import re
 import time
 
-from cryptography.hazmat.backends import default_backend
-from cryptography import x509
-from cryptography.x509 import SignatureAlgorithmOID as algos
-
-from .http import HTTPStatus
-
 import requests
+from cryptography import x509
+from cryptography.hazmat.backends import default_backend
+from cryptography.x509 import SignatureAlgorithmOID as AlgOID
 
-from .common import MIME_JSON, TokenInfo, Authentication, CommonConnection, KeyType, Policy, ZoneConfig, CertField
+from .common import MIME_JSON, TokenInfo, Authentication, KeyType, Policy, ZoneConfig, CertField
+from .connection_tpp_abstract import AbstractTPPConnection, URLS
 from .errors import (ClientBadData, ServerUnexptedBehavior, AuthenticationError, CertificateRequestError,
                      CertificateRenewError)
+from .http import HTTPStatus
 from .pem import parse_pem
 
 HEADER_AUTHORIZATION = "Authorization"  # type: str
@@ -43,27 +42,7 @@ KEY_REFRESH_TOKEN = "refresh_token"  # type: str  # nosec
 KEY_EXPIRATION_DATE = "expiration_date"  # type: str
 
 
-class URLS:
-    API_TOKEN_URL = "vedauth/"  # type: str  # nosec
-    API_BASE_URL = "vedsdk/"  # type: str  # nosec
-
-    AUTHORIZE_TOKEN = API_TOKEN_URL + "authorize/oauth"  # type: str
-    REFRESH_TOKEN = API_TOKEN_URL + "authorize/token"  # type: str
-    REVOKE_TOKEN = API_TOKEN_URL + "revoke/token"  # type: str
-
-    AUTHORIZE = API_BASE_URL + "authorize/"
-    CERTIFICATE_REQUESTS = API_BASE_URL + "certificates/request"
-    CERTIFICATE_RETRIEVE = API_BASE_URL + "certificates/retrieve"
-    FIND_POLICY = API_BASE_URL + "config/findpolicy"
-    CERTIFICATE_REVOKE = API_BASE_URL + "certificates/revoke"
-    CERTIFICATE_RENEW = API_BASE_URL + "certificates/renew"
-    CERTIFICATE_SEARCH = API_BASE_URL + "certificates/"
-    CERTIFICATE_IMPORT = API_BASE_URL + "certificates/import"
-    ZONE_CONFIG = API_BASE_URL + "certificates/checkpolicy"
-    CONFIG_READ_DN = API_BASE_URL + "Config/ReadDn"
-
-
-class TPPTokenConnection(CommonConnection):
+class TPPTokenConnection(AbstractTPPConnection):
     def __init__(self, url, user=None, password=None, access_token=None, refresh_token=None, http_request_kwargs=None):
         """
         :param str url:
@@ -164,10 +143,31 @@ class TPPTokenConnection(CommonConnection):
                 request_data["CASpecificAttributes"].append(ca_origin)
             else:
                 request_data["CASpecificAttributes"] = [ca_origin]
+
+        if request.custom_fields:
+            custom_fields_map = {}
+            for c_field in request.custom_fields:
+                if custom_fields_map.get(c_field.name):
+                    custom_fields_map[c_field.name].append(c_field.value)
+                else:
+                    custom_fields_map[c_field.name] = [c_field.value]
+
+            for key in custom_fields_map:
+                custom_field_json = {
+                    "Name": key,
+                    "Values": custom_fields_map[key]
+                }
+                if request_data.get("CustomFields"):
+                    request_data["CustomFields"].append(custom_field_json)
+                else:
+                    request_data["CustomFields"] = [custom_field_json]
+
         status, data = self._post(URLS.CERTIFICATE_REQUESTS, data=request_data)
         if status == HTTPStatus.OK:
             request.id = data['CertificateDN']
-            log.debug("Certificate sucessfully requested with request id %s." % request.id)
+            request.cert_guid = data['Guid']
+            log.debug("Certificate successfully requested with request id %s." % request.id)
+            log.debug("Certificate successfully requested with GUID %s." % request.cert_guid)
             return True
 
         log.error("Request status is not %s. %s." % HTTPStatus.OK, status)
@@ -263,8 +263,8 @@ class TPPTokenConnection(CommonConnection):
                         upns.append(x.value[2 : :])
                 request.user_principal_names = upns
                 request.uniform_resource_identifiers = list([x.value for x in e.value if isinstance(x,x509.UniformResourceIdentifier)])
-        if cert.signature_algorithm_oid in (algos.ECDSA_WITH_SHA1, algos.ECDSA_WITH_SHA224, algos.ECDSA_WITH_SHA256,
-                                            algos.ECDSA_WITH_SHA384, algos.ECDSA_WITH_SHA512):
+        if cert.signature_algorithm_oid in (AlgOID.ECDSA_WITH_SHA1, AlgOID.ECDSA_WITH_SHA224, AlgOID.ECDSA_WITH_SHA256,
+                                            AlgOID.ECDSA_WITH_SHA384, AlgOID.ECDSA_WITH_SHA512):
             request.key_type = (KeyType.ECDSA, KeyType.ALLOWED_CURVES[0])
         else:
             request.key_type = KeyType(KeyType.RSA, 2048)  # todo: make parsing key size
@@ -321,6 +321,7 @@ class TPPTokenConnection(CommonConnection):
             key_type = KeyType(KeyType.ECDSA, data["Policy"]["KeyPair"]["EllipticCurve"]["Value"])
         else:
             key_type = None
+
         z = ZoneConfig(
             organization=CertField(s['Organization']['Value'], locked=s['Organization']['Locked']),
             organizational_unit=CertField(ou, locked=s['OrganizationalUnit']['Locked']),
@@ -365,11 +366,8 @@ class TPPTokenConnection(CommonConnection):
             raise ClientBadData("Certificate not found by thumbprint")
         return data['Certificates'][0]['DN']
 
-    def _read_config_dn(self, dn, attribute_name):
-        status, data = self._post(URLS.CONFIG_READ_DN, {
-            "ObjectDN": dn,
-            "AttributeName": attribute_name,
-        })
+    def _get_certificate_details(self, cert_guid):
+        status, data = self._get(URLS.CERTIFICATE_SEARCH + cert_guid)
         if status != HTTPStatus.OK:
             raise ServerUnexptedBehavior("")
         return data
